// Class   : ConcurrentServer
// Author  : Russ Ethington
// Version : 8/5/97 for Java 1.1

import java.util.*;
import java.net.*;
import java.io.*;

public class ConcurrentServer implements Runnable, Cloneable
{
   private Thread            life = null;
   private Thread            twin = null;
   private ServerSocket      server = null;
   private Socket            socket = null;

   private InputStream       in = null;
   private OutputStream      out = null;

   private boolean           original = true;
   private boolean           runState = false;
   private boolean           connectState = false;

   private int               connectionID = -1;
   private static Vector     connectionList = null;

   final int                 defaultPort = 4646;

   public static void main(String args[])
   {
      new ConcurrentServer().start();
      System.out.println("Ready.");
   }

   public ConcurrentServer () { }

   public void start() { start(defaultPort); }

   public void start(int thePort)
   {
      connectionList = new Vector();

      if (life == null)
      {
         life = new Thread(this);
         life.setPriority(Thread.MAX_PRIORITY);
      }
      try
      {
         server = new ServerSocket(thePort);
      }
      catch (IOException e)
      {
         trouble("Could not create new server socket..." + e);
         return;
      }
      life.start();
   }

   public void stop()
   {
      if (life != null)
         life.stop();
      shutdown();
   }

   public boolean isRunning() { return runState; }

   public boolean isConnected() { return connectState; }

   public int getConnectionID() { return connectionID; }

   public void run()
   {
      if (original) // this is the primary server thread.
      {
         runState = true;
         while (true)
         {
            try
            {
               socket = server.accept();
            }
            catch (IOException e)
            {
               trouble("Could not accept incoming calls...");
               shutdown();
               return;
            }

            ConcurrentServer copy = null;
            try
            {
               copy = (ConcurrentServer) clone();
            }
            catch (CloneNotSupportedException e)
            {
               trouble("Unexpected 'CloneNotSupported' exception..." + e);
            }

            copy.life = null;
            copy.server = null;
            copy.original = false;
            copy.runState = false;
            copy.addToConnectionList();
            twin = new Thread(copy);
            twin.start();
         }
      }
      else // this is the twin clone coming to life!
      {
         runState = true;
         connectState = true;
      
         handleSession(socket);
         shutdown();
      }
   }

   protected void addToConnectionList()
   {
      int found = -1;
      synchronized (connectionList)
      {
         for (int i = 0; i < connectionList.size(); i++)
         {
            if (connectionList.elementAt(i) == null)
            {
               found = i;
               break;
            }
         }
         if (found >= 0)
         {
            connectionID = found;
            connectionList.setElementAt(this, found);
         }
         else
         {
            connectionID = connectionList.size();
            connectionList.addElement(this);
         }
      }
   }

   protected void removeFromConnectionList()
   {
      connectionList.setElementAt(null, connectionID);
      connectionID = -1;
   }

   public int activeConnections()
   {
      int count = 0;
      synchronized (connectionList)
      {
         for (int i = 0; i < connectionList.size(); i++)
            if (connectionList.elementAt(i) != null)
               count++;
      }
      return count;
   }

   public ConcurrentServer getConnection(int i)
   {
      synchronized (connectionList)
      {
         if (i < connectionList.size())
            return (ConcurrentServer) connectionList.elementAt(i);
         else
            return null;
      }
   }

   public void shutdown()
   {
      removeFromConnectionList();

      life = null;
      twin = null;
      runState = false;
      if (connectState)
      {
         try
         {
            if ( in != null) in.close();
            if ( out != null) out.close();
            
         }  catch (IOException e) { System.err.println(e); }
         connectState = false;
         if (socket != null)
         {
            try
            {
               socket.close();
            }
            catch (Exception e) {System.err.println(e); }
         }
      }
      socket = null;
      server = null;

      in = null;
      out = null;
   }

   protected void trouble(String message)
   {
      System.out.println("ConcurrentServer: " + message);
   }

   public void handleSession(Socket socket)
   {
      System.out.println("ConcurrentServer:handleSession()");

      // This should be overridden in a subclass.
   }
}
