/************************************************************
*          This is a part of the VideoClick project.
*               Copyright (C) 2000 Vsoft Ltd.
*                   All rights reserved.
* NAME: RefCountingRot.cpp
* DESCRIPTION:
* Singelton object keeping track of the refcounts for all interface
* for the full lifetime of an application.
*
* AUTHOR: Noam Cohen.
*
* $Archive:  will be filled by sourceSafe  $
* $Revision:  will be filled by sourceSafe $
* $Author:  will be filled by sourceSafe    $
* $JustDate:   will be filled by sourceSafe $
************************************************************/

#include "StdAfx.h"
#include "DbgUtils.h"

#include "RefCountingRot.h"
#include "INamedObject2.h"

#ifndef SAFE_FREE
#define SAFE_FREE(x)	if(x){ free(x); x = NULL;}
#endif

ObjParams::ObjParams()
{
	functionName = NULL;
	line = 0;
	szFile = NULL;
}

// copy ctor is needed to force deep copy
ObjParams::ObjParams(const ObjParams& obj)
{
	// although strdup() can handle null argument, BoundsChecker complains...
	if(NULL != obj.functionName)
		functionName = strdup(obj.functionName);
	else
		functionName = NULL;

	line = obj.line;

	if(NULL != obj.szFile)
		szFile = strdup(obj.szFile);
	else
		szFile = NULL;
}

ObjParams::~ObjParams()
{
	SAFE_FREE(functionName);
	line = 0;
	SAFE_FREE(szFile);
}

RefCountingRot::RefCountingRot()
{
	TRACE("RefCountingRot created\n");
}

// when I die, all COM object must already be dead.
// Report if it is not the state.
RefCountingRot::~RefCountingRot()
{
	ObjectsMap::iterator iter;
	
	// for each object in the map
	for(iter = m_ObjMap.begin(); iter != m_ObjMap.end(); iter++)
	{
		ObjData &objParams = (*iter).second;
		if(objParams.addRefed.size() != objParams.released.size()  )
		{
			// mismatch between AddRef and Release calls
			DumpDataFor(iter);
		} 
	}

}


// ==================== RefCountingRot::Register ====================
// 		interface AddRef() => register this call.
// HISTORY:
//		Date		Intls.	Notes
//		--------	------	------------------------------------
//		27/07/00	Noam	Created
void RefCountingRot::Register(void* pThis, ObjParams &objParam)
{
	ObjIterator iter;

	iter = m_ObjMap.find(pThis);
	// if there is no entry for this object, make it now.
	if( iter == m_ObjMap.end() )
	{ 
		// create new entry in the map
		ObjData objData;
		std::pair<ObjIterator, bool> res = m_ObjMap.insert(ObjectsMap::value_type(pThis, objData));
		iter = res.first;

		// Try to put some informative data for this object.
		// the data is taken by quering INameObject2
		SetObjectData(pThis, &iter->second);
	}

	int i;
	ParamList &theList = ((*iter).second).addRefed;
	i = theList.size();
	//ParamList::iterator itr;
	theList.insert( theList.end(), objParam); 
	i = theList.size();
}



// ==================== RefCountingRot::UnRegister ====================
// 	report interface Release
//	It is possible that calls to Release and hence UnRegister() will precede any AddRef,
//	therefore we take care of it with caution.
// HISTORY:
//		Date		Intls.	Notes
//		--------	------	------------------------------------
//		27/07/00	noam	Created
void RefCountingRot::UnRegister(void* pThis, ObjParams &objParam)
{
	ObjIterator iter;
	iter = m_ObjMap.find(pThis);
	//todo: handle a "key not found" error
	if( iter == m_ObjMap.end() )
	{ 
		// create new entry in the map
		ObjData objData;
		std::pair<ObjIterator, bool> res = m_ObjMap.insert(ObjectsMap::value_type(pThis, objData));
		iter = res.first;
	}

	ParamList &theList = ((*iter).second).released;
	theList.insert( theList.end(), objParam); 
}

void RefCountingRot::DumpDataFor(ObjIterator & it)
{
	CStdString msg , objName;
	void* pObj = (*it).first;
	ObjData &objParams = (*it).second;

	// print all the available data in nice format.
	ObjectName & rObjName = objParams.objName;
	if(rObjName.m_bIsValid)
	{
		objName.Format("{%s(%s)}", (const char*)rObjName.m_ClassName, (const char*)rObjName.m_ObjName);
	}
	else
		objName.Format("<this == 0x%X>",reinterpret_cast<unsigned int>(pObj));

	msg.Format("Object %s has %d AddRef and %d Release called\n",(const char*) objName,
		objParams.addRefed.size(),
		objParams.released.size() );
	msg += CStdString("\nAddRef:\n");
	DumpParamList(msg,objParams.addRefed);
	msg += CStdString("\n------------\nRelease:\n");
	DumpParamList(msg,objParams.released);
	msg += CStdString("\n------------\n\n\n");

	MessageBox(NULL,msg,"Reference leaks",MB_ICONWARNING | MB_OK);

}

void RefCountingRot::DumpParamList(CStdString & msg, ParamList & aList)
{
	ParamList::iterator i;
	for(i= aList.begin(); i != aList.end(); i++)
	{
		ObjParams &record = *i;
		CStdString t;
		if(record.line == 0)
		{// no data available for this record
			t.Format("???\n");
		}
		else
			t.Format("%s , %d , %s\n", record.szFile , record.line, record.functionName);
		msg += t;
	}
}


// ==================== RefCountingRot::SetObjectData ====================
// try getting some data from the COM object and put it in the ROT
// HISTORY:
//		Date		Intls.	Notes
//		--------	------	------------------------------------
//		11/10/2000	Noam	Created
void RefCountingRot::SetObjectData(void* pThis, ObjData* pObjData)
{
	IUnknown* pUnk = (IUnknown*)pThis; // most dangerous act - may cause access violation at later stage if the assumption is wrong!
	INamedObject2 * pName;
	HRESULT hr = pUnk->QueryInterface(IID_INamedObject2, (void**) & pName);
	if(SUCCEEDED(hr))
	{ 
		// get the data 
		pObjData->objName.m_ClassName = pName->GetClassName();
		pObjData->objName.m_bIsValid = true;
		pName->Release();
	}
	else
	{ // interface not supported
	}

}
