/************************************************************
*          This is a part of the VideoClick project.
*               Copyright (C) 1998 Vsoft Ltd.
*                   All rights reserved.
* NAME: RefCountingRot.h
* DESCRIPTION:
*  The Reference Counting Running Object Table  maintains
*  the current registration of AddRef/Release calls.
*  This data is used when the destructon of the class is called
* to create a report of unmatching calls to Register/Unregister
*
* AUTHOR: Noam Cohen
*
* $Archive:  will be filled by sourceSafe  $
* $Revision:  will be filled by sourceSafe $
* $Author:  will be filled by sourceSafe    $
* $JustDate:   will be filled by sourceSafe $
************************************************************/

#ifndef RefCountingRot_H
#define RefCountingRot_H

#include "WarningsOff.h"
#include <map>
#include <list>
#include "StdString.h"
#include "WarningsOn.h"


/* Objects of this class are used to identify a single function caller.
  Each call record is identified by <line number, file, function name>
*/
class ObjParams
{
public:
	ObjParams();
	
	//copy ctor
	ObjParams(const ObjParams& obj); 

	virtual ~ObjParams();

// data made public for simplicity.

	int line;			// source file line number
	char* szFile;		// file name e.g "main.cpp"
	char* functionName; // name of function calling the callee. e.g.  "MyClass::SetParam()"
}; 


// The object name struct contains identifying info on a COM object.
// Note: Currently, it is collected by asking the object ( INamedObject2)
class ObjectName
{
public:
	ObjectName()
	{
		m_bIsValid = false;
	}

// data made public for simplicity.

	CStdString	m_ClassName;	// friendly name of the class e.g. "ouput pin"
	CStdString	m_ObjName;		// friendly name of a class instance e.g. "pin(0xE0)"
	GUID	m_CLSID;			// GUID of the class. May be equal to CLSID of the COM object.
	GUID	m_OBJID;			// GUID of this specific instance. May be left unused.
	bool	m_bIsValid;			// true if values of the object are initialized.
};


/*
This class maintains the current registration of AddRef/Release calls.

This data is used when the destructon of the class is called
to create a report of unmatching calls to Register/Unregister.

Ideally, This class is a singleton, but it is more convinient to have one
for each COM object.

The internal data structure:
Each object is a key in a map ( using the 'this' value), called m_ObjMap
The value is a struct with one vector for AddRef() and one for Release().
Each vector contains the registration of the calls made to Register/Unregister.
A single vector element is <int line number, char[] file name, char[] function name>

*/
class RefCountingRot
{
public:
    RefCountingRot();
    ~RefCountingRot();

    /*------------------------------------------------------------------------
	Register a call to AddRef() for a certain object.
	The called object is identified by pThis.
	The caller is identified by objParam.
	Each called object has its own list of AddRef/Release calls
    */
    void Register( void* pThis, 	// [IN] - the COM object whose AddRef() is called
                   ObjParams &objParam  	// [IN] - parameters of the caller
                 );

    /*------------------------------------------------------------------------
	Register a call to Release() for a certain object.
	Note: if Register() for pThis was not called before, no action is taken.
    */
    void UnRegister( void* pThis, 	// [IN] - the COM object whose Release() is called
                     ObjParams &objParam  	// [IN] -  parameters of the caller
                   );

private:
	// the ParamList is the ordered container of the params
	typedef std::list<ObjParams> ParamList;

	// the ObjData contains all the data relevant to one key ( the interface pointer)
	typedef struct { 
		ObjectName objName;
		ParamList addRefed;
		ParamList released;
	}  ObjData;

	// the container of all the <interface pointer, value> associations
	typedef std::map<void*, ObjData> ObjectsMap;

	// iterator for the container
	typedef ObjectsMap::iterator ObjIterator;

	void DumpDataFor(ObjIterator & it);

	// Dump the data of a single list into msg.
	void DumpParamList(CStdString & msg, ParamList & aList);

	void SetObjectData(void* pThis, ObjData* pObjData);

	ObjectsMap m_ObjMap;

};


#endif